@@ -42,18 +42,17 @@
 # Requirements
 # ============
 # sudo apt install apache2 libapache2-mod-perl2
-# sudo apt install qemu-kvm bridge-utils qemu iptables dnsmasq libguestfs-tools
+# sudo apt install qemu-system bridge-utils qemu iptables dnsmasq libguestfs-tools
 # sudo apt install novnc websockify
 #
 # sudo apt install libsys-cpu-perl libsys-cpuload-perl libsys-meminfo-perl \
 #   libcgi-pm-perl liblist-moreutils-perl libnet-dns-perl libjson-perl \
 #   libproc-background-perl libproc-processtable-perl libemail-valid-perl \
 #   libnet-smtps-perl libmail-imapclient-perl libnet-ldap-perl libemail-valid-perl \
-#   libwww-perl liburi-perl libnet-ssh2-perl
-# optional: libtext-qrcode-perl
+#   libwww-perl liburi-perl libnet-ssh2-perl libtext-qrcode-perl
 #
 # (c) 2020- Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.
-# https://gitlab.com/soleil-data-treatment/remote-desktop
+# https://gitlab.com/soleil-data-treatment/soleil-software-projects/qemu-web-desktop
 
 
 
@@ -74,24 +73,25 @@ use Cwd 'abs_path';     # for abs_path
 use File::Temp      qw/ tempdir tempfile /;
 use File::Path      qw/ rmtree  /;
 use File::Basename  qw/ fileparse dirname /;
+use File::Copy      qw/ copy/;
 use List::MoreUtils qw(uniq); # liblist-moreutils-perl
 use List::Util;
-use Sys::CPU;           # libsys-cpu-perl           for CPU::cpu_count
-use JSON;               # libjson-perl              for JSON
 use IO::Socket::INET;
 use IO::Socket::IP;
-use Sys::MemInfo;
+use JSON;               # libjson-perl              for JSON
 use Proc::Background;   # libproc-background-perl   for Background->new
 use Proc::ProcessTable; # libproc-processtable-perl
 use Proc::Killfam;      # libproc-processtable-perl for killfam (kill pid and children)
+use Sys::CPU;           # libsys-cpu-perl           for CPU::cpu_count
+use Sys::MemInfo;
 
-use Net::SMTPS;         # libnet-smtps-perl         for smtp user check and emailing
-use Mail::IMAPClient;   # libmail-imapclient-perl   for imap user check
-use Net::LDAP;          # libnet-ldap-perl          for ldap user check
-use Net::SSH2;          # libnet-ssh2-perl          for local user check
 use Email::Valid;       # libemail-valid-perl
+use Mail::IMAPClient;   # libmail-imapclient-perl   for imap user check
 use LWP::UserAgent;     # libwww-perl               for new and get(URL)
 use Net::Ping;          # perl                      to test server up-down
+use Net::SMTPS;         # libnet-smtps-perl         for smtp user check and emailing
+use Net::LDAP;          # libnet-ldap-perl          for ldap user check
+use Net::SSH2;          # libnet-ssh2-perl          for local user check
 use URI;                # liburi-perl               to split URL into parts
 
 #   Sys::CpuLoad          libsys-cpuload-perl       for CpuLoad::load
@@ -122,7 +122,7 @@ use constant IS_CGI      => IS_MOD_PERL || exists $ENV{'GATEWAY_INTERFACE'};
 # we use a Hash to store the configuration. This is simpler to pass to functions.
 our %config;
 
-$config{version}                  = "25.02.28";  # year.month.day
+$config{version}                  = "25.04.05";  # year.month.day
 
 my $dirname = abs_path(dirname(__FILE__));
 
@@ -132,7 +132,7 @@ if (-e "$dirname/../config.pl") {
 } elsif (-e '/etc/qemu-web-desktop/config.pl') {
   require '/etc/qemu-web-desktop/config.pl';
 } else {
-  die "Can not find config.pl (e.g. /etc/qemu-web-desktop/config.pl or $dirname/../config.pl";
+  die "$0: Can not find config.pl (e.g. /etc/qemu-web-desktop/config.pl or $dirname/../config.pl";
 }
 
 # search detached GPU (via vfio-pci). Only use video part, no audio.
@@ -170,10 +170,12 @@ for(my $i = 0; $i < @ARGV; $i++) {
     foreach my $key (sort keys %config) {
       print STDERR "  --$key=VALUE [$config{$key}]\n";
     }
+    print STDERR "\nTo manually start virtual machine in DIR_VM/VM :\n";
+    print STDERR "$0 --dir_machines=DIR_VM --machine=VM --dir_snapshots=/tmp --dir_cfg=/tmp --oneshot=1 --service_max_mem_fraction_nb_per_user=0.8 --service_max_cpu_fraction_nb_per_user=0.8 --snapshot_alloc_disk=30\n";
     exit;
   } elsif (/--version|-v$/) {
     print STDERR "$0 $config{version}\n\n";
-    print STDERR "Copyright (C) 2020-     Synchrotron Soleil,     GPL3.\n";
+    print STDERR "Copyright (C) 2020-     Synchrotron Soleil, AGPL3.\n";
     print STDERR "This is free software; see the source for copying conditions.  There is NO\n";
     print STDERR "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n";
     print STDERR "Written by Emmanuel Farhi <emmanuel.farhi\@synchrotron-soleil.fr>\n";
@@ -191,7 +193,7 @@ for(my $i = 0; $i < @ARGV; $i++) {
 
 $config{session_nb}               = 0;
 
-if ($config{session_stop}) {
+if ($config{session_stop}) { # from ARGS (command line)
   # wait for session to end, and clean files/PIDs.
   my $session_ref = session_load(\%config, $config{session_stop});
   if ($session_ref) {
@@ -228,7 +230,7 @@ my %session;
 $session{machine}     = $config{machine};
 # OpenStack DEFAULT.instance_name_template ?
 $session{dir_snapshot}= tempdir(TEMPLATE => "$config{service}" . "_XXXXXXXX", 
-  DIR => $config{dir_snapshots}) || die "Can not create snapshot directory in $config{dir_snapshots}.";
+  DIR => $config{dir_snapshots}) || die "$0: Can not create snapshot directory in $config{dir_snapshots}.";
 $session{name}        = File::Basename::fileparse($session{dir_snapshot});
 $session{snapshot}    = "$session{dir_snapshot}/$config{service}.qcow2";
 $session{json}        = "$config{dir_cfg}/$session{name}.json";
@@ -241,7 +243,6 @@ $session{persistent}  = "";  # not persistent implies lower server load
 $session{cpu}         = $config{snapshot_alloc_cpu};  # default: cores
 $session{memory}      = $config{snapshot_alloc_mem};  # default: in GB
 $session{disk}        = $config{snapshot_alloc_disk}; # default: only for ISO
-$session{video}       = $config{qemu_video};          # default: driver to use
 $session{gpu}         = ""; # indicates PCI GPU passthrough request when not empty
 $session{config_script}= ""; # a file/URL to execute before starting the VM
 $session{snapshot_lifetime} = $config{snapshot_lifetime}; # default [sec]
@@ -266,6 +267,66 @@ $session{token_file}  = "";
 $session{runs_as_cgi} = IS_CGI;
 $session{url}         = "";
 
+# auto: default QEMU executable
+if ($config{qemu_exec} =~ /^(auto|host)$/) {
+  my $arch=`uname -m`;
+  if    ($arch =~ 'x86_64')          { $config{qemu_exec} = 'qemu-system-x86_64';  }
+  elsif ($arch =~ /(aarch64|arm64)/) { $config{qemu_exec} = 'qemu-system-aarch64'; }
+  else { die "$0: Can not determine qemu-system for architecture $arch. Set config{qemu_exec} in config.pl\n"; }
+}
+# auto: best QEMU options from host capabilities
+if ($config{qemu_exec_options} =~ 'auto') {
+  $config{qemu_exec_options} =~ s/auto//;
+  my $accel=`$config{qemu_exec} -accel help`;
+  my $mach=`$config{qemu_exec} -M help`;
+  
+  # test type of accelerator
+  # usual case on Intel/AMD and Apple/Silicon
+  if ($accel =~ "kvm") { 
+    $config{qemu_exec_options} .= " -enable-kvm -cpu host";
+    if ($mach =~ "Standard PC") { $config{qemu_exec_options} .= ",kvm=off"; }
+  } elsif ($accel =~ "hvf") { $config{qemu_exec_options} .= " -accel hvf  -cpu host";  }
+
+  # test type of target machine
+  if ($mach =~ "Standard PC") {
+    # usual case on Intel/AMD "PC"
+    $config{qemu_exec_options} .= " -M pc";
+    # test type of VGA (not on ARM)
+    my $mach_vga= `$config{qemu_exec} -M pc -vga help`;
+    if    ($mach_vga =~ "virtio") { $config{qemu_exec_options} .= " -vga virtio"; }
+    elsif ($mach_vga =~ "qxl")    { $config{qemu_exec_options} .= " -vga qxl";    }
+  } elsif ($mach =~ "virt") { 
+    # Other, e.g. ARM
+    $config{qemu_exec_options} .= " -M virt,iommu=smmuv3"; 
+  }
+
+# the device virtio-gpu is nice, but brings mouse pointer offset
+#  if (`$config{qemu_exec} -device help` =~ "virtio-gpu") {
+#    $config{qemu_exec_options} .= " -device virtio-gpu";
+#  }
+
+  # special case for ARM: require EFI boot
+  if (`uname -m` =~ /(aarch64|arm64)/) {
+    # search for EFI/ARM
+    my @EFIs=('/usr/share/qemu-efi-aarch64/QEMU_EFI.fd',
+              '/usr/share/qemu-efi/QEMU_EFI.fd',
+              '/usr/share/ovmf/AARCH64/QEMU_EFI.fd',
+              '/usr/share/edk2/arm/QEMU_EFI.fd',
+              '/opt/homebrew/share/qemu/edk2-aarch64-code.fd');
+    foreach my $EFI (@EFIs) {
+      if (-f $EFI) {
+        # Debian arm
+        #   https://documentation.ubuntu.com/server/how-to/virtualisation/qemu/index.html
+        #   https://wiki.debian.org/Arm64Qemu recommends to use the '-bios $EFI'
+        # Special case for Apple/Silicon
+        #   https://ww2.coastal.edu/mmurphy2/oer/qemu/aarch64/
+        $config{qemu_exec_options} .= " -bios $EFI";
+        last;
+      }
+    }
+  }
+} # end auto qemu options
+
 # ------------------------------------------------------------------------------
 # Update session info from CGI
 # ------------------------------------------------------------------------------
@@ -702,12 +763,11 @@ if (not $error) {
   # common options for QEMU
   my $cmd = "$config{qemu_exec} -smp $session{cpu} "
     . " -name $session{name}:$session{user}:$session{machine}"
-    . " -machine pc,accel=kvm -enable-kvm -cpu host,kvm=off"
+    . " $config{qemu_exec_options}"
     . " -m $session{memory} -device virtio-balloon"
     . " -hda $session{snapshot} -device ich9-ahci,id=ahci"
-    . " -netdev user,id=mynet0 -device virtio-net,netdev=mynet0"
-    . " -usb -device usb-tablet"
-    . " -vga $session{video}";
+    . " -netdev user,id=mynet0 -device virtio-net,netdev=mynet0 -usb"
+    . " -device qemu-xhci -device usb-kbd -device usb-tablet";
     
   # add copy/paste support (not always functional)
   $cmd .= " -chardev qemu-vdagent,id=ch1,name=vdagent,clipboard=on -device virtio-serial-pci"
@@ -763,7 +823,7 @@ if (not $error) {
     print $token_handle "change vnc password\n$session{token_vnc}\n";
     close($token_handle);
     # redirect 'token' to QEMU monitor STDIN to set the VNC password
-    $cmd .= ",password -monitor stdio > /dev/null < $token_name";
+    $cmd .= ",password=on -monitor stdio > /dev/null < $token_name";
   } else {
     $cmd .= ' > /dev/null';
   }
@@ -777,6 +837,7 @@ if (not $error) {
   if (not $proc_qemu or not $proc_qemu->pid) {
     $error  .= "Could not start QEMU/KVM for $session{machine}.\n";
   } else {
+    print STDERR $0.": $session{name}: START: $cmd\n";
     if ($config{layout_level} >= 2) {
       $output .= "<li>$ok Started QEMU/KVM for $session{machine} with VNC.</li>\n";
     }
@@ -859,13 +920,17 @@ if (not $error) {
       $session{pid_wait} = $proc_novnc->pid;
     }
   }
-  
+  # build the session URL
+  my $append=$config{service_reverse_proxy}; # may be "" or ":port" or "/alias"
+  if ($append eq "" or $config{service_port_multiple}) {
+    $append=":$session{port}"; # default
+  }
   if ($config{service_port_multiple}) {
-    $session{url} = "$config{http}://$session{server_name}:$session{port}/vnc.html?resize=scale&autoconnect=true&host=$session{server_name}&port=$session{port}";
+    $session{url} = "$config{http}://$session{server_name}$append/vnc.html?resize=scale&autoconnect=true&host=$session{server_name}&port=$session{port}";
   } else {
     # in case of an Apache2 port redirection, change ':$session{port}' into e.g. 
     # the Apache2 Location (e.g. '/darts').
-    $session{url} = "$config{http}://$session{server_name}:$session{port}/vnc.html?resize=scale&autoconnect=true&path=?token=$session{token_vnc}&port=$session{port}";
+    $session{url} = "$config{http}://$session{server_name}$append/vnc.html?resize=scale&autoconnect=true&path=?token=$session{token_vnc}&port=$session{port}";
   }
 }
 
@@ -1040,7 +1105,7 @@ if (not $error and $proc_novnc and $proc_qemu and $session{oneshot}) {
 	session_stop(\%session);
 }
 
-
+exit;
 
 
 
@@ -1491,10 +1556,8 @@ sub service_purge {
   
   foreach my $snapshot (glob("$dir/$service"."_*")) {
       if (-d $snapshot) { # is a snapshot directory
-        print STDERR $0." [$session{date}] service_purge: snapshot: $snapshot\n";
         my $snaphot_name = fileparse($snapshot); # just the session name
         if (-e "$cfg/$snaphot_name.json") {
-          print STDERR $0." [$session{date}] service_purge: json found\n";
           my $session_ref = session_load(\%config, "$cfg/$snaphot_name.json");
           session_stop($session_ref);
         }
